#!/usr/bin/perl -w

# Yoonha Choi   June/29/07
# Modified : Feb/15/10  - Fixed the bug at Genotype 6 (line 236-250)
#
# Purpose
# -------
# Estimate k-statistics(k0,k1,k2) for every pair of individuals
#
# Usage
# -----
# kstat.pl <freq file> <genotype file> <output file>
#
# e.g.) kstat.pl set0.frq set0.dat set0.out
#
#
# Input Files
# -----------
# 1. allele frequency data :e.g.)set0.frq
# 2. genotype data of every individual: e.g.)set0.dat
# 3. output file : e.g.)set0.out
#
# Format of input files:
# - freq data
#   : mrk_name    allele    count   freq
#     rs??????      C        120    1.0000
#     ...          ...       ...    ...
# - genotype data
#   : subject_ID   mrk1     mrk2     mrk3 ... mrkN
#      NA????      C/A       T/T     T/G  ...  G/G
###
#
# Output Files
# -----------
# the name of output file is given by a user.
# e.g.)set0.out
#
# format:   ind1    ind2     k0    k1     k2
#           NA????  NA????   0.00  1.00  0.00
#           ...     ...      ...   ...   ...
#
######

########## ARGUEMENTS ##########
if (@ARGV<3){
  print "\nUSAGE: ./kstat.pl <freq file> <genotype file> <output file>\n";
  print "EXAMPLE: ./kstat.pl set0.frq set0.dat set0.out\n";
  exit(0);
}

($frqfile,$gntfile,$outfile)=@ARGV;       #Files Name


######### INPUT FRQ ###########
open(FREQ, "<$frqfile")or die "couldn't open $frqfile $! \n";
$header = <FREQ>;
$line = <FREQ>;

$marker_number = 0;
($mrk_name,$allele,$tmp,$frq) = split(" ",$line);
$prev_mrk = $mrk_name;
push(@{$freq_allele[$marker_number]},$allele,$frq);

while(<FREQ>){
  ($mrk_name,$allele,$tmp,$frq) = split(" ",$_);

  if($prev_mrk ne $mrk_name){
    $marker_number++;
  }
  push(@{$freq_allele[$marker_number]},$allele,$frq);
  $prev_mrk = $mrk_name;
}


########## INPUT GENOTYPE ##########
open(GNT,"<$gntfile") or die "Couldn't open $gntfile \n";
$header = <GNT>;
($tmp,@mrk_name) = split(" ",$header);

$N_mrk = $#mrk_name;

$ind_i = 0;
while(<GNT>){
  ($ind,@{$mrk_gnt[$ind_i]})=split(" ",$_);
  push(@sub_id, $ind);
  $ind_i = $ind_i + 1;
}
$N_ind = ($ind_i -1);

####### Print Information #######
print ("The number of individuals is $ind_i \n");
printf ("The number of markers is %5d \n",($N_mrk+1));


########## MAIN Routine #########
open(OUT,">$outfile") or die "Couldn't open $outfile \n";
printf OUT ("%10s %10s %7s %7s %7s %10s %10s\n","sub_1","sub_2","k0","k1","k2","N_mrk","N_itr");

#open(LOG,">kstat.log") or die "Coudln't open kstat.log \n";

foreach $i (0..($N_ind - 1)){
  foreach $j (($i+1)..$N_ind){

#foreach $i (0..0){
#  foreach $j (1..1){

    $sub1 = $sub_id[$i];
    $sub2 = $sub_id[$j];

    @gnt1 = @{$mrk_gnt[$i]};
    @gnt2 = @{$mrk_gnt[$j]};

    $N_obs = 0;
    foreach $g (0..$N_mrk){
      if(($gnt1[$g] ne "-0-") &&($gnt2[$g] ne "-0-")){

	############## Set Allele Frequencies ##############
	($obs11,$obs12)=split("/",$gnt1[$g]);
	($obs21,$obs22)=split("/",$gnt2[$g]);

	@frq_tmp = @{$freq_allele[$g]};
	%frequency = @frq_tmp;

	if(exists $frequency{$obs11}){
	  $frq11 = $frequency{$obs11};
	}else{$frq11 = 0;}
	if(exists $frequency{$obs12}){
	  $frq12 = $frequency{$obs12};
	}else{$frq12 = 0;}
	if(exists $frequency{$obs21}){
	  $frq21 = $frequency{$obs21};
	}else{$frq21 = 0;}
	if(exists $frequency{$obs22}){
	  $frq22 = $frequency{$obs22};
	}else{$frq22 = 0;}

	## Status of Homozygote and Number of shared alleles ##
	$homo1 = 0;
	$homo2 = 0;
	if($obs11 eq $obs12){$homo1 = 1;}
	if($obs21 eq $obs22){$homo2 = 1;}

	## Seven cases of Genotype and Computation of P(G|X) ##
	$geno_i = 0;
	$p0[$N_obs] =0;
	$p1[$N_obs] =0;
	$p2[$N_obs] =0;
	$G_type[$N_obs]=0;
	###### Homozygote/ Homozygote ######
	if($homo1==1 && $homo2==1){
	  if($obs11 eq $obs21){     ### Genotype 1
	    $pi = $frq11;
	    $p0[$N_obs] = $pi**4;   # P(G|X=0)
	    $p1[$N_obs] = $pi**3;   # P(G|X=1)
	    $p2[$N_obs] = $pi**2;   # P(G|X=2)
	    $geno_i =1;
	    $G_type[$N_obs]=1;
	  }
	  if($obs11 ne $obs21){     ### Genotype 2
	    $pi = $frq11;
	    $pj = $frq21;
	    $p0[$N_obs] = ($pi**2)*($pj**2);    # P(G|X=0)
	    $geno_i =1;
	    $G_type[$N_obs]=2;
	  }
	}

	###### Homozygote/ Heterozygote ######
	if($homo1 ==1 && $homo2 ==0){ ### Genotype 3
	  if($obs11 eq $obs21 ){
	    $pi = $frq11;
	    $pj = $frq22;
	    $p0[$N_obs] = 2*($pi**3)*($pj);
	    $p1[$N_obs] = ($pi**2)*($pj);
	    $geno_i =1;
	    $G_type[$N_obs]=3;
	  }elsif($obs11 eq $obs22){
	    $pi = $frq11;
	    $pj = $frq21;
	    $p0[$N_obs] = 2*($pi**3)*($pj);
	    $p1[$N_obs] = ($pi**2)*($pj);
	    $geno_i =1;
	    $G_type[$N_obs]=3;
	  }elsif($obs11 ne $obs21 && $obs11 ne $obs22){ ### Genotype 4
	    $pi = $frq11;
	    $pj = $frq21;
	    $pm = $frq22;
	    $p0[$N_obs] = 2*($pi**2)*($pj)*($pm);
	    $geno_i =1;
	    $G_type[$N_obs]=4;
	  }
	}
	if($homo1 ==0 && $homo2 ==1){
	  if($obs11 eq $obs21 ){    ### Genotype 3
	    $pi = $frq21;
	    $pj = $frq12;
	    $p0[$N_obs] = 2*($pi**3)*($pj);
	    $p1[$N_obs] = ($pi**2)*($pj);
	    $geno_i =1;
	    $G_type[$N_obs]=3;
	  }elsif($obs12 eq $obs21){
	    $pi = $frq21;
	    $pj = $frq11;
	    $p0[$N_obs] = 2*($pi**3)*($pj);
	    $p1[$N_obs] = ($pi**2)*($pj);
	    $geno_i =1;
	    $G_type[$N_obs]=3;
	  }elsif($obs21 ne $obs11 && $obs21 ne $obs12){  ### Genotype 4
	    $pi = $frq21;
	    $pj = $frq11;
	    $pm = $frq12;
	    $p0[$N_obs] = 2*($pi**2)*($pj)*($pm);
	    $geno_i =1;
	    $G_type[$N_obs]=4;
	  }
	}

	###### Heterozygote/ Heterozygote ######
	if($homo1 ==0 && $homo2 ==0){
	  $share = 0;
	  if($obs11 eq $obs21 || $obs11 eq $obs22){
	    $share++;
	  }
	  if($obs12 eq $obs21 || $obs12 eq $obs22){
	    $share++;
	  }
	  if($share == 2){    ### Genotype 5
	    $pi = $frq11;
	    $pj = $frq12;
	    $p0[$N_obs] = 4*($pi**2)*($pj**2);
	    $p1[$N_obs] = ($pi)*($pj)*($pi+$pj);
	    $p2[$N_obs] = 2*($pi)*($pj);
	    $geno_i = 1;
	    $G_type[$N_obs]=5;
	  }
	  if($share == 1){    ### Genotype 6
	    $pi = $frq11;
	    $pj = $frq12;
	    if($obs21 ne $obs11 && $obs21 ne $obs12){
	      $pm = $frq21;
	      $pc = $frq22;
	    }else{
	      $pm = $frq22;
	      $pc = $frq21;
	    }
	    $p0[$N_obs] = 4*($pi)*($pc)*($pj)*($pm);
	    $p1[$N_obs] = ($pi)*($pj)*($pm);
	    $geno_i= 1;
	    $G_type[$N_obs]=6;
	  }
	  if($share == 0){     ### Genotype 7
	    $p0[$N_obs] = 4*($frq11)*($frq12)*($frq21)*($frq22);
	    $geno_i = 1;
	    $G_type[$N_obs]=7;
	  }
	} #The end of "	if($homo1 ==0 && $homo2 ==0){"

	if($geno_i ne 1){
	  print "Error : There is no genotype which matches to observation\n";
	}

#printf LOG ("%10s %10s %10s\n",$gnt1[$g],$gnt2[$g],$G_type[$N_obs]);
	$N_obs++;


      }# the end of "	if(($gnt1[$g] ne "-0-") &&($gnt2[$g] ne "-0-")){"
    } # the end of " foreach $g (0..$N_mrk){"

#print ("the number of used markers is $N_obs\n");

    ############# EM Algorithm #################
    $diff = 1;     #Initial Values
    $k0 = 0.4;
    $k1 = 0.3;
    $k2 = 0.3;

    $N_itr=0;
    while($diff > 0.0001){
      $N_itr++;

      ### E-Step ###
      $k0_pre = $k0;
      $k1_pre = $k1;
      $k2_pre = $k2;

      $k0_num=0;
      $k1_num=0;
      $k2_num=0;
      foreach $obs (0..($N_obs-1)){
	$p_total = ($k0 * $p0[$obs])+ ($k1 * $p1[$obs]) +($k2 * $p2[$obs]);
	if($p_total ==0){
	  $X0[$obs] = 0;
	  $X1[$obs] = 0;
	  $X2[$obs] = 0;
	}else{
	  $X0[$obs] = ($k0*$p0[$obs])/$p_total;
	  $X1[$obs] = ($k1*$p1[$obs])/$p_total;
	  $X2[$obs] = ($k2*$p2[$obs])/$p_total;
	  $k0_num = $k0_num+$X0[$obs];
	  $k1_num = $k1_num+$X1[$obs];
	  $k2_num = $k2_num+$X2[$obs];
	}
      }
      $k0 = $k0_num/$N_obs;
      $k1 = $k1_num/$N_obs;
      $k2 = $k2_num/$N_obs;

      $diff = abs($k0-$k0_pre) + abs($k1-$k1_pre) + abs($k2-$k2_pre);

    }# the end of "    while($diff > 0.0001){"
    printf OUT ("%10s %10s %.5f %.5f %.5f %10s %10s\n", $sub1,$sub2,$k0,$k1,$k2,$N_obs,$N_itr);

# Checking
if($i eq 45 && $j eq 46){ print "sub45 and sub46\n";}

 } # the end of "foreach $j (($i+1)..$N_ind){"
}# the end of "  foreach $i (0..($N_ind -1)){
